var net = require('net');

var JsonSocket = require('json-socket');
var Promise = require('bluebird');
var _ = require('lodash');

module.exports = Client;

/// Clinet
function Client(options) {
    this.options = _.defaults(options || {}, {
        host: '127.0.0.1',
        autoReconnect: false
    });
    this._closed = true;
    this._queue = [];
    this.resetReconnectTimeout();
}

/**
 * Send message to daemon and get messages from daemon
 * @param port
 * @param host
 * @param data
 * @optional
 * @returns {Promise}
 */
Client.sendMessage = function(port, host, data) {
    return new Promise(function(resolve, reject) {
        if (arguments.length === 2 || typeof host !== 'string') {
            data = host;
            host = undefined;
        }

        var socket = new JsonSocket(new net.Socket());

        socket.connect(port, host);

        socket.on('connect', function(message){
            socket.sendEndMessage(data);
            resolve(message);
        });

        socket.on('error', function(err){
            socket.end();
            reject(err);
        });
    });
};
/**
 * create simple socket
 * @param port
 * @param host
 * @returns {JsonSocket}
 */
Client.connect = function(port, host) {
    var socket = new JsonSocket(new net.Socket());
    socket.connect(port, host);
    return socket;
};

Client.prototype.__proto__ = require('events').EventEmitter.prototype;


Client.prototype.reset = function (port, host){
    clearTimeout(this.reconnectTO);

    if (port) {
        this.options.port = port;
    }

    if (host) {
        this.options.host = host;
    }
};

Client.prototype.resetReconnectTimeout = function(){
    this.reconnectTimeout = 500;
};


Client.prototype.end = function (){
    this.socket.end();
};

Client.prototype.connect = function (){
    var port = this.options.port;
    var host = this.options.host;

    if (!port || !host) {
        this.emit('error', new Error('no port or no host'));
        return;
    }

    if (this.socket) {
        this.socket.end();
    }

    var socket = new JsonSocket(new net.Socket());

    this.socket = socket;

    socket.on('close', function() {
        this._closed = true;
        if (this.options.autoReconnect) {
            this.reconnect();
        } else {
            this.emit('close');
        }
    }.bind(this));

    socket.on('end', function(){
        this._closed = true;
        this.emit('end');
    }.bind(this));

    socket.on('error', function(err) {
        this._closed = true;
        if (!this.options.autoReconnect) {
            this.emit('error', err);
        }
    }.bind(this));

    socket.on('connect', function() {
        this._closed = false;
        this.resetReconnectTimeout();
        this.emit('connect', this);

        var queue = this._queue;
        if (this._queue.length) {
            this._queue = [];
            queue.forEach(function(msg){
                socket.sendMessage(msg);
            });
        }
    }.bind(this));

    socket.on('message', function(message) {
        if (message.type === 'request') {
            if (_.isNumber(message._callback)) {
                this.emit('request', message.channel, message.options, function(err, result){
                    socket.sendMessage({
                        type: 'callback',
                        _callback: message._callback,
                        args: [err, result]
                    });
                });
            }
            return;
        }

        if (message.success === 'false' && message.error) {
            console.error('server error %s', message.error);
            return;
        }

        if(message.type) {
            this.emit('type:' + message.type, message);
        } else {
            this.emit('message', message);
        }

    }.bind(this));

    socket.connect(port, host);
};

Client.prototype.reconnect = function (){
    clearTimeout(this.reconnectTO);

    this.reconnectTO = setTimeout(function(){

        this.connect();
        var maxTimeout = 20 * 60 * 1000; //5min

        if (this.reconnectTimeout < maxTimeout) {
            this.reconnectTimeout = this.reconnectTimeout * 2;
        }

        if (this.reconnectTimeout > maxTimeout) {
            this.reconnectTimeout = maxTimeout;
        }

    }.bind(this), this.reconnectTimeout);
    this.emit('reconnect', this.reconnectTimeout);
};

Client.prototype.send = function (data) {

    if (this._closed) {
        this._queue.push(data);
        this.reconnect();
    } else {
        this.socket.sendMessage(data);
    }

};