// Node
var path = require('path');

// NPM
var _ = require('lodash');
var glob = require('glob');
var Promise = require('bluebird');

// Project
var dUtil = require('util-def');
var pkg = require('../package.json');

var argv = require('yargs')
    .alias('v', 'verbose')
    .boolean('verbose')
    .boolean('noColor')
    .boolean('daemon')
    .default('noColor', false)
    .argv;

var def = module.exports = {
    version: pkg.version
};

var PluginRuntime  = require('./plugin-runtime')(def);

/**
 * 命令行模块
 * @type {cli|exports}
 */
def.cli = require('./cli');


/**
 * Log模块
 * @type {cli|exports}
 */
def.log = dUtil.log;

if (argv.verbose || argv.v) {
    def.log.level = -1;
} else if (argv.silent) {
    def.log.level = 'silent';
}

if (argv.noColor) {
    def.log.disableColor();
    dUtil.color.enabled = false;
    def.noColor = true;
}

if (argv.daemon) {
    def.childMod = true;
}

if (argv._[0] === 'setup') {
    def.mode = 'setup';
}

/**
 * DEF 根目录
 * @type {cli|exports}
 */
def.home = process.env['DEF_ROOT'] || dUtil.system.getUserHome('.def');

def.log.verbose('def.home', def.home);
def.log.verbose('def.version', def.version);

/**
 * 软件根目录
 * @type {*}
 */
def.softHome = path.join(def.home, 'soft');

/**
 * 统一配置管理
 * @type {db|exports}
 */
def.config = require('./config');

/**
 * DEF 初始化检查模块
 * @type {exports}
 */
def.init = require('./init');

/**
 * 发送统计信息模块
 * @type {sendMessage}
 */

_.extend(def, require('./analytics'));

/**
 * 插件对象
 * @type {Object}
 * @private
 */
def._plugins = {};

/**
 * 通一的错误处理
 * @type {function}
 */
def.fatal = function onError(err, from){
    this.log.error(err, from);


    if (err && err.stack && this.log.level < 1) {
        this.log.error(err.stack)
    } else {
        this.log.info('tips', '运行 def %s -v 查看详情', process.argv.splice(2).join(' '));
    }

    process.exit(1);
};

/**
 * 解析 request 目标
 * @type {function}
 * @param {String} action request target
 */
def.parseAction = function parseAction(action) {
    var a = action.split(':');

    if (a.length >=2 && !a[0]) {
        a[0] = this.name;
    }
    return {
        plugin: a.shift(),
        method: a.shift(),
        args: a
    }
};

/**
 * get plugin runtime by name, awp-> @ali/def-awp or  awp-> def-awp
 * @param name
 * @returns {PluginRuntime}
 */
def.getPlugin = function (name) {
    var plugin = this._plugins[name];

    if (!plugin) {
        return;
    }

    plugin.load();
    return plugin;
};

/**
 * call plugin api
 * @param action
 * @param {Object} data params of plugin
 * @param {PluginRuntime} from
 * @returns {Promise}
 */
def.request = function(action, data, from) {

    var caller;
    var p = def.parseAction(action);
    var plugin = p.plugin;
    var method = p.method;

    if (!from) {
        caller = 'core';
    } else {
        caller = from.name;
    }

    this.log.verbose('request', action + ' from:' + caller);

    var resolver = Promise.defer();

    var target = def.getPlugin(plugin);

    if (target) {
        var fn = target.published[method];

        if (fn) {

            var runInfo = {
                from: caller
            };

            if (fn.length >= 3) {
                fn.call(target, data, runInfo, resolver.callback);
            } else {
                var ret = fn.call(target, data, runInfo);
                if (Promise.is(ret)) {
                    return ret;
                } else {
                    resolver.resolve(ret);
                }
            }
        } else {
            resolver.reject('plugin \'' + plugin + '\' has no method \'' + method + '\'');
        }

    } else {
        def.log.verbose('request', 'plugin not found, try to load');

        try {

            var possiblePluginPath = [];
            possiblePluginPath.push(path.join(def.home, 'node_modules/@ali/', 'def-' + plugin));
            possiblePluginPath.push(path.join(def.home, 'node_modules', 'def-' + plugin));

            while(possiblePluginPath.length) {
                def.registerPlugin(possiblePluginPath.pop());
            }

            return def.request.apply(def, arguments);
        } catch (e) {
            def.log.verbose('request', 'faild to  load plugin');
            resolver.reject(e);
        }

        resolver.reject('plugin \'' + plugin + '\' no found!');
    }

    return resolver.promise;
};

if (process.platform === 'win32' || process.platform === 'darwin') {
    def.windows = require('./service');
}


/**
 * Register plugin to DEF
 * @param where
 */
def.registerPlugin = function(where) {
    var pkg = require(path.join(where, 'package.json'));
    var runtime = new PluginRuntime(where, pkg);

    def.log.verbose('plugin', 'register plugin: %s', dUtil.color.cyan(runtime.name + '@' + pkg.version));

    def._plugins[runtime.name] = runtime;
};

def.checkDependencies = function(){
    def = this;

    return new Promise(function(resolve, reject){
        var allDependencies = [];

        _.mapValues(def._plugins, function(plugin){
            allDependencies = allDependencies.concat(plugin.pkg.defDependences ||[]);
        });

        _.pairs(def._plugins).reduce(function(allDependencies, current){
            var plugin = current[1];
            if (plugin.pkg && plugin.pkg.defDependences) {
                allDependencies = allDependencies.concat(plugin.pkg.defDependences);
            }
            return allDependencies;
        }, []);

        var missing = _(allDependencies).uniq().remove(function(name){
            return _.has(def._plugins, name);
        }).value();

        if (!missing.length) {
            resolve()
        } else {
            reject(new Error('Plugin dependencies broken'));
        }
    });

};

def.loadPlugins = function loadPlugins() {
    var def = this;

    return new Promise(function(resolve){
        var pluginDir = path.join(def.home, 'node_modules');

        var modules = ['def-*/package.json', '@ali/def-*/package.json'].reduce(
            function(memo, item){
                return memo.concat(glob.sync( item,{cwd: pluginDir}));
            }, []);

        modules
            .map(function(module) {
                return path.dirname(module);
            })
            .filter(function(module) {
                return !/^def-cli/.test(module);
            })
            .forEach(function(dirname) {
                var where = path.join(pluginDir, dirname);

                try {
                    def.registerPlugin(where);
                } catch(e) {
                    def.log.error('fail to load plugin: ' + dirname + ', use --verbose to show more info' );
                    def.log.verbose(e);
                    def.log.verbose(e.stack);
                }
            });
        resolve();
    });

};

def.run = function () {
    def
        .init()
        .then(def.loadPlugins.bind(def))
        .then(def.sendCommand.bind(def))
        .then(function(){
            _.mapValues(def._plugins, function(plugin){
                def.cli.program.command(plugin.name, plugin.description);
            });
        })
        .then(def.cli.bind(def, argv))
        .catch(def.fatal);
};

process.on('uncaughtException', function(err) {
    if(def && def.fatal) {
        def.fatal(err, 'uncaughtException');
    }
});