/**
 * 1. 检查并初始化 .def 目录
 * 2. 检查或安装 预装插件
 * 3. 初始化DEF配置
 * @type {exports}
 */


var mkdirp = require('mkdirp');
var path = require('path');
var fs = require('fs');
var glob = require('glob');
var _ = require('lodash');
var Promise = require('bluebird');
var dUtil = require('util-def');
var semver = require('semver');

/**
 * init def
 * @returns {Promise}
 */
var init = module.exports = function init() {

    this.log.verbose('init', 'start');

    //初始化DEF配置
    this.config.init(path.join(this.home, 'conf.json'));

    return initHome.bind(this)()
        .then(initPkgJSON)
        .then(initInfo)
        .then(checkPlugin.bind(this))
        .then(checkUpdate.bind(this))
        .then(checkNode.bind(this));
};


var initHome = init.initHome = function() {
    var def = this;
    return new Promise(function(resolve) {
        var dir = def.home;
        if (fs.existsSync(dir) && fs.statSync(dir).isFile()) {
            fs.unlinkSync(dir);
        }

        if (!fs.existsSync(dir)) {
            mkdirp.sync(dir);
        }
        resolve(def);
    });

};

var initPkgJSON = init.initPkgJSON = function(def){
    return new Promise(function(resolve) {
        var home = def.home;
        var pkgJSONFile = path.join(home, 'package.json');

        if (!fs.existsSync(pkgJSONFile)) {
            fs.writeFileSync(pkgJSONFile, JSON.stringify({
                "name": "def-home",
                "version": "0.0.0",
                "private": true
            }, null, 4));
        }
        resolve(def);
    });

};


/**
 * 从Git中获取用户邮箱
 * @returns {*}
 */
function getInfo() {
    var gitConfigFile = dUtil.system.getUserHome('.gitconfig');
    if (!fs.existsSync(gitConfigFile)) {
        return {}
    }
    var file = fs.readFileSync(gitConfigFile, 'utf8');
    var nameMatch = file.match(/name\s+=\s+(\S+)\s*$/m);
    var emailMatch = file.match(/email\s+=\s+(\S+)\s*$/m);
    return {
        email: emailMatch?emailMatch[1] : '',
        name: nameMatch? nameMatch[1] : ''
    }
}

/**
 * 初始化用户信息
 * @returns {*}
 */
var initInfo = init.initInfo = function initInfo(def) {
//    def.log.verbose('init', 'user info');

    return new Promise(function(resolve, reject) {
        if (def.config.get('email') && def.config.get('mid')) {
            resolve(def);
            return;
        }

        if (!def.config.get('mid')) {
            var mid = 'M' + Date.now() + ':' + Math.round(100000 * Math.random());
            def.config.set('mid', mid);
        }

        var info = getInfo();

        if (info.email) {
            def.config.set('email', info.email);
        }

        if (info.name) {
            def.config.set('name', info.name);
        }

        def.config.save().then(function(){
            resolve(def);
        }, reject);
    });
};

/**
 * 检查预装插件
 * @returns {Promise}
 */
function checkPlugin(def) {

    var preInstallPlugin = [
        'def-cli',
        'def-plugin',
        'def-alias'
    ];

    def.log.verbose('init', 'check preInstall Plugins');
    return new Promise(function(resolve, reject){
        // 扫描，安装预装插件
        glob('*', {
            cwd: path.join(def.home, 'node_modules')
        }, function(err, matches) {
            var missing = _.without.apply(_, [preInstallPlugin].concat(matches));

            if (missing.length) {
                def.log.info('installing... ' + missing.join(', '));
                dUtil.install(missing, def.home).then(function(){
                    resolve(matches.concat(missing));
                }, reject);
            } else {
                resolve(def);
            }
        });
    });
}

///**
// * get current checkout from package.json
// * @param {Object} pkg
// * @param {String} pkg.version
// * @param {String} pkg.readme
// * @param space
// * @returns {String}
// */
//function getChangeLog(pkg, space) {
//    var input = pkg.readme;
//    var version = pkg.version;
//    var regTxt = '###.*'+ version.replace(/\./g, '\\.' ) + '.*\\n';
//    var reg = new RegExp(regTxt, 'g');
//    var out = '';
//    var result = input.split(reg);
//
//    if (result.length < 2) {
//        return out;
//    }
//
//    result = result[1].split(/#+/)[0].trim();
//
//    if (result) {
//        out = result.split('\n').map(function(line) {
//            return line.replace(/^\s*(-|\+|\*)\s*(.+)$/, getSpace(space) + '+ $2');
//        }).join('\n');
//    }
//
//    return out;
//}

function getSpace(length) {
    var out = '';

    while(length) {
        out += ' ';
        length--;
    }
    return out;
}

var dailyCheck = init.dailyCheck = function (){

};

function checkUpdate(modules) {
    var def = this;
    return new Promise(function(resolve) {
        def.log.verbose('check update', 'checking...');
        var lastUpdate = new Date(def.config.get('lastCheck') || 0);
        var now = new Date();
//        var expired = true;
        var expired = (
            now.getFullYear() !== lastUpdate.getFullYear()
            || now.getMonth() !== lastUpdate.getMonth()
            || now.getDate() !== lastUpdate.getDate());

        if (def.childMod) {
            def.log.verbose('check update', 'exit by childMode');
            resolve(modules);
            return;
        }

        if (!expired) {
            def.log.verbose('check update', 'exit by not expired');
            resolve(modules);
            return;
        }

        def.log.info('check update', '正在检查插件更新...');
        def.config.set('lastCheck', Date.now());

        function done() {
            def.config.save().then(function(){
                resolve(modules);
            });
        }

        def.request('plugin:update', {})
            .then(function(){
                def.log.info('check update', '自动更新已经完成');
                done();
            })
            .catch(function(){
                done();
            });
    });
}


function checkNode() {
    var def = this;

    return new Promise(function(resolve){

        var requireVersion = '0.10.21';
        var currentVersion = process.version;

        if (semver.lt(currentVersion, requireVersion)) {
            def.log.warn('check', 'Nodejs 版本检查:');
            def.log.warn('check', '当前版本（%s）低于 (%s)', currentVersion, requireVersion);
            def.log.warn('check', '请升级到最新版本 Nodejs http://nodejs.org/');
            def.log.warn('check', '当前版本，某些模块可能无法安装 (报错)');
            setTimeout(resolve, 3000);
            return;
        }
        resolve();
    });
}
